@@ -12,11 +12,12 @@ with contextlib.redirect_stdout(None):
     import pygame
     import pygame.gfxdraw
 import curses
+from functools import cache
 
 from perlcompat import die, warn, getopts
 import tbdump
 
-FONT_SIZE = 18
+FONT_SIZE = 14
 
 BLACK = (0, 0, 0)
 
@@ -26,6 +27,7 @@ GRAY = 101
 WHITE = 102
 
 # This code is taken from cellx/monitor/color.py.
+@cache
 def hsv2rgb(h, s, v):
     """Convert color in HSV (Hue, Saturation, Value) space to RGB.  Return (R,
     G, B) as a list.  Transformation algorithm is taken from
@@ -66,7 +68,7 @@ class Series(list):
             label = f'#{id_}'
         self.label = label
         self.hide = False
-        self.vsum = 0	# Used for obtaining the sample mean.
+        self.vsum = 0  # Used for obtaining the sample mean.
 
     def __repr__(self):
         return f'Series(id_={self.id_}, vmin={self.vmin}, vmax={self.vmax}, color={self.color}, label={self.label}, hide={self.hide}, len={len(self)}'
@@ -80,19 +82,15 @@ class Series(list):
         self.vsum += v
         super().append(v)
 
-    def at(self, ratio, window=None):
-        """For a given window size WINDOW, pick a value located at the ratio
-        of RATIO in all samples withn the last WINDOW.  If window is not
-        specified, take a sample from all data."""
-        if window is None:
-            n = int(len(self) * ratio)
-            return self[n]
-        # If the number of samples is not enough, pick from all available samples.
-        if len(self) < window:
-            return self.at(ratio, None)
-        # Otherwise, pick from the window of samples at the end.
-        offset = len(self) - window
-        n = int(offset + window * ratio)
+    def at(self, ratio, left, right):
+        """Return a value located at the ratio of RATIO in samples between
+        LEFT and RIGHT.  If LEFT or RIGHT is negative, the offset is regarded
+        from the end."""
+        if left < 0:
+            left = max(len(self) + left, 0)
+        if right < 0:
+            right += len(self)
+        n = int(left + (right - left) * ratio)
         return self[n]
 
 class Plot:
@@ -101,6 +99,8 @@ class Plot:
                  width=None,
                  height=None,
                  offset=None,
+                 left=0,
+                 right=-1,
                  grid=None,
                  subgrid=None,
                  start_color=0):
@@ -115,11 +115,12 @@ class Plot:
         if offset is None:
             offset = 0, 0
         self.offset = offset
+        self.left = left
+        self.right = right
         self.grid = grid
         self.subgrid = subgrid
         self.start_color = start_color
         self._series = []
-        self.window = None
         self.vmin = None
         self.vmax = None
 
@@ -149,7 +150,8 @@ class Plot:
         """Assuming that the y-axis ranges from the minimum and the maximum
         value, return the y-axsis of the value V.  For instance, if V is eqaul
         to the minimum, this function returns the value of SELF.HEIGHT."""
-        return int(self.height - self.height * (v - self.vmin) /
+        # NOTE: -1 is for fitting within the screen.
+        return int((self.height - 1) - (self.height - 1) * (v - self.vmin) /
                    (self.vmax - self.vmin))
 
     def draw_single_series(self, sr):
@@ -158,7 +160,7 @@ class Plot:
         x0, y0 = None, None
         for x in range(self.width):
             ratio = x / self.width
-            v = sr.at(ratio, self.window)
+            v = sr.at(ratio, self.left, self.right)
             y = self.to_y_coordinate(v)
             if x > 0:
                 self.screen.draw_line(x0,
@@ -197,14 +199,14 @@ class Plot:
         for n, sr in enumerate(self._series):
             x = 1
             y = 1 + n * FONT_SIZE
-            v = sr[-1]
+            v = sr[self.right]
             # Display field label.
             self.screen.draw_text(x, y, sr.label, sr.color, offset=self.offset)
             # Display current, mean, and maximum values.
             mean = sr.vsum / len(sr)
-            self.screen.draw_text(x + FONT_SIZE * 4,
+            self.screen.draw_text(x + FONT_SIZE * 3,
                                   y,
-                                  f'{v:8.2f}/{mean:8.2f}/{sr.vmax:8.2f}',
+                                  f'{v:9.2f} AVG{mean:9.2f} MAX{sr.vmax:9.2f}',
                                   sr.color,
                                   offset=self.offset)
 
@@ -226,32 +228,43 @@ class Plot:
         # Draw legends.
         self.draw_legends()
         # Display the maximum and the minimum values.
-        self.screen.draw_text(self.width - FONT_SIZE * 6,
+        self.screen.draw_text(self.width - FONT_SIZE * 5,
                               0,
                               f'{self.vmax:8.2f}',
-                              WHITE,
+                              self.start_color,
                               offset=self.offset)
-        self.screen.draw_text(self.width - FONT_SIZE * 6,
+        self.screen.draw_text(self.width - FONT_SIZE * 5,
                               self.height - FONT_SIZE * 2,
                               f'{self.vmin:8.2f}',
-                              WHITE,
+                              self.start_color,
                               offset=self.offset)
 
-    def draw_background(self):
-        """Fill the background with gradient colors."""
-        # FIXME: Should not re-generate at every drawing.
+    @cache
+    def create_background(self):
+        surface = pygame.Surface((self.width, self.height))
         for y in range(self.height):
-            alpha = 48 - int(48 * y / self.height)
+            alpha = int(48 * y / self.height)
             self.screen.draw_line(0,
                                   y,
                                   self.width,
                                   y,
                                   self.start_color,
-                                  offset=self.offset,
-                                  alpha=alpha)
+                                  alpha=alpha,
+                                  surface=surface)
+        return surface
+
+    def draw_background(self):
+        """Fill the background with gradient colors."""
+        # FIXME: Should not re-generate at every drawing.
+        bg = self.create_background()
+        self.screen.screen.blit(bg, self.offset)
 
 class Screen:
-    def __init__(self, curses=False, width=None, height=None):
+    def __init__(self,
+                 curses=False,
+                 width=None,
+                 height=None,
+                 fullscreen=False):
         self.curses = curses
         # Uses SVGA (800x600 pixels) window by default.
         if not curses and (width is None or height is None):
@@ -264,6 +277,7 @@ class Screen:
     def __repr__(self):
         return f'Screen(curses={self.curses}, width={self.width}, height={self.height})'
 
+    @cache
     def color_rgba(self, n, alpha=255):
         """Return 8-bit RGBA color for the color number N as a tuple.  The
         alpha channel is given by ALPHA."""
@@ -299,7 +313,15 @@ class Screen:
             curses.init_pair(7, curses.COLOR_WHITE, curses.COLOR_BLACK)
             self.screen.erase()
 
-    def draw_line(self, x1, y1, x2, y2, color=0, offset=None, alpha=255):
+    def draw_line(self,
+                  x1,
+                  y1,
+                  x2,
+                  y2,
+                  color=0,
+                  offset=None,
+                  alpha=255,
+                  surface=None):
         """Draw a straight line connecting two points: (X1, Y1) and (X2, Y2).
         The line color can b especified by COLOR and ALPHA."""
         if offset is None:
@@ -308,8 +330,10 @@ class Screen:
         x1, y1 = x1 + dx, y1 + dy
         x2, y2 = x2 + dx, y2 + dy
         if not self.curses:
+            if surface is None:
+                surface = self.screen
             rgba = self.color_rgba(color, alpha)
-            pygame.gfxdraw.line(self.screen, x1, y1, x2, y2, rgba)
+            pygame.gfxdraw.line(surface, x1, y1, x2, y2, rgba)
             return
 
         # Curses only.
@@ -339,7 +363,18 @@ class Screen:
                 x = int(x1 + (x2 - x1) * (y - y1) / (y2 - y1))
                 self.draw_text(x, y, point, color)
 
-    def draw_text(self, x, y, text, color=0, size=FONT_SIZE, offset=None):
+    @cache
+    def load_font(self, name, size, bold=False):
+        return pygame.font.SysFont(name, size, bold=bold)
+
+    def draw_text(self,
+                  x,
+                  y,
+                  text,
+                  color=0,
+                  size=FONT_SIZE,
+                  offset=None,
+                  _cache=[]):
         """Display a text TEXT at the location of (X, Y).  Font size and color
         can be specified with COLOR and SIZE."""
         if offset is None:
@@ -347,8 +382,7 @@ class Screen:
         dx, dy = offset
         x, y = x + dx, y + dy
         if not self.curses:
-            # FIXME: Loaded font object should be re-utilized.
-            font = pygame.font.SysFont('Courier', size, bold=True)
+            font = self.load_font('Courier', size, bold=True)
             rgba = self.color_rgba(color)
             # The second parameter True means enabling antialiasing.
             text = font.render(text, True, rgba, BLACK)
@@ -384,3 +418,16 @@ class Screen:
         else:
             # Not necessary in curses mode.
             pass
+
+    def scan_key(self):
+        if not self.curses:
+            # Pump event as much as possible.
+            while True:
+                event = pygame.event.poll()
+                if event.type == pygame.NOEVENT:
+                    return None
+                if event.type == pygame.KEYDOWN:
+                    return event.unicode
+        else:
+            # Not necessary in curses mode.
+            pass
